import stamp.core.*;
import stamp.math.*;

import stamp.peripheral.memory.eeprom.*;                // Contains handling for EEPROM operations
import stamp.peripheral.sensor.GPS.GPSMath.*;           // Contains the GPS calculations
import stamp.peripheral.sensor.GPS.NMEA.*;              // Contains the NMEA messages and other NMEA specific things
import stamp.peripheral.sensor.GPS.StringOperations.*;  // Contains some string handling classes
import stamp.peripheral.sensor.GPS.Include.*;           // Contains the linked list class and other stuff
import stamp.peripheral.sensor.GPS.Keyboard.*;          // Contains the memkey implementation
import stamp.peripheral.sensor.GPS.Temperature.*;       // Contains the SHT11 sensor wrapper
import stamp.peripheral.display.lcd.CrystalFontz.*;     // Contains the LCD layer
import stamp.peripheral.sensor.GPS.DisplayDriver.*;     // Contains the LCD driver for the CrystalFontz
import stamp.peripheral.sensor.humidity.sht11.*;        // The temperature layer
import stamp.peripheral.sensor.GPS.Kernel.*;            // The SailSmart kernel
import stamp.peripheral.sensor.GPS.RouteFollower.*;     // Contains the class handling keeping track of were we are and where to go
import stamp.peripheral.sensor.GPS.Menu.*;              // Contains the GPS menu

public class SailSmart
{

  public static void main()
  {
    // STEP 1:
    // Create the display driver
    DisplayDriver displayDriver = new DisplayDriver(new CrystalFontz(new Uart( Uart.dirTransmit,
                                                                               CPU.pin10,
                                                                               Uart.dontInvert,
                                                                               Uart.speed19200,
                                                                               Uart.stop1  )),
                                                    new CrystalFontz(new Uart( Uart.dirTransmit,
                                                                               CPU.pin11,
                                                                               Uart.dontInvert,
                                                                               Uart.speed19200,
                                                                               Uart.stop1  )));


    // STEP 1.x
    // Do some display initialization
    {
      // Hide the cursor
      displayDriver.ShowCursor(DisplayDriver.DISPLAY_1 | DisplayDriver.DISPLAY_2, false);

      // We do not allow scrolling
      displayDriver.ScrollOff(DisplayDriver.DISPLAY_1 | DisplayDriver.DISPLAY_2, true );

      // Clear the display
      displayDriver.ClearScreen(DisplayDriver.DISPLAY_1 | DisplayDriver.DISPLAY_2);

    }


    //displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 0, "Initializing..." );


    // STEP 2:
    // Create the global class storing the SHARED variables used in this program
    GlobalVars globalVars = null;
    try
    {
      globalVars = new GlobalVars(displayDriver);
    }
    catch(Exception ex)
    {
      // If the uM FPU is not OK, do not continue
      displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 2, 0, "*");
      return;
    }

    // STEP 3:
    // The route handler
    //displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 2, 0, "Route handler-");
    MemoryRouteHandler routeHandler = new MemoryRouteHandler(CPU.pin7, CPU.pin8, globalVars);
    if (!routeHandler.IsFormatted())
    {
      routeHandler.Format();
      displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 3, 4, 0, 20 );
    }
    //displayDriver.WriteMessage( DisplayDriver.DISPLAY_1, 2, 16, routeHandler.IsFormatted() ? "OK!" : "Error!");

    if (!routeHandler.IsFormatted()) return;

     /*
    // ----> Added
    MEMKey keyboard = new MEMKey();
    displayDriver.WriteMessage( DisplayDriver.DISPLAY_2, 1, 16, "OK!");

    RouteFollower routeFollower = new RouteFollower(routeHandler);
    GPSMenuHandler menuHandler = new GPSMenuHandler( displayDriver, globalVars, routeHandler, keyboard, routeFollower );
    while ( true )
    {
      menuHandler.RunMenu();
    }
    */
    // <------


    // STEP 4:
    // The GPS receive UART
    //displayDriver.WriteMessage( DisplayDriver.DISPLAY_1, 3, 0, "GPS - init...");

    Uart rxGPS = new Uart(Uart.dirReceive, CPU.pin12, Uart.dontInvert, Uart.speed4800, Uart.stop1);
    //displayDriver.WriteMessage( DisplayDriver.DISPLAY_1, 3, 16, "OK!");

    // STEP 5:
    // The GPS math is doing all the gps calculations - like course, distance and other things
    //displayDriver.WriteMessage( DisplayDriver.DISPLAY_2, 0, 0, "GPS services...");
    GPSMath gpsMath = new GPSMath(globalVars);
    RouteFollower routeFollower = new RouteFollower(routeHandler);

    // STEP 6:
    // Contains the GPS message lists
    NMEA0183 nmea0183 = new NMEA0183(rxGPS, globalVars);
    //displayDriver.WriteMessage( DisplayDriver.DISPLAY_2, 0, 16, "OK!");

    // STEP 7:
    // Interface to the MEMKey
    //displayDriver.WriteMessage( DisplayDriver.DISPLAY_2, 1, 0, "Keyboard...");

    MEMKey keyboard = new MEMKey();
    //displayDriver.WriteMessage( DisplayDriver.DISPLAY_2, 1, 16, "OK!");

    // STEP 8:
    // The temperature sensor
    //displayDriver.WriteMessage( DisplayDriver.DISPLAY_2, 2, 0, "Temp sensor...");


    sht11 sht = new sht11(CPU.pin5, CPU.pin6);
    Temperature temp = new Temperature(sht);

    //displayDriver.WriteMessage( DisplayDriver.DISPLAY_2, 2, 16, "OK!");

    // STEP 9:
    // The kernel is actually running this application
    //displayDriver.WriteMessage( DisplayDriver.DISPLAY_2, 3, 0, "Kernel start...");

    // STEP 10:
    // The GPS Menu lets the user handle the GPS menues and interact with this program
    Kernel GPSKernel = new Kernel(displayDriver, globalVars, routeHandler, nmea0183, keyboard, temp, routeFollower, gpsMath);

    // Let the kernel go wild
    GPSKernel.Go();
    /**/

  }
}