package stamp.peripheral.memory.eeprom;

import stamp.core.*;
import stamp.peripheral.sensor.GPS.Include.*;          // Contains the linked list class and other stuff
import stamp.peripheral.sensor.GPS.DisplayDriver.*;
import com.micromegacorp.math.hardware.*;

/**

This class contains basic functionality for handling the MC24LC256 eeprom memory

*/
public class MemoryRouteHandler
{
  public static final int m_dev = 0;                                 // Device ID (default zero)
  private MC24LC256 m_eeprom;                       // memory handling base class
  //private DisplayDriver m_displayDriver = null;   // Handle to the display driver class
  private static final int m_eepromSize = 32767;    // Total size of the eeprom (32768 -> 0-32767)
  private GlobalVars m_globalVars;                  // This class contains global variables used in the whole program

  // Please see the specifications below
  public static final int FOMMAT_TABLE_START = 0;
  public static final int SETTINGS_START = 5;
  public static final int ROUTE_TABLE_START = 1009;
  public static final int ROUTES_START = 1765;
  public static final int MEMORY_USAGE_END = 32005;


  /** **** Memory map (in bytes)****
  FORMAT INFO:  0 - 4           -> Formatted table information
  SETTINGS:     5 - 1008        -> Contains settings and other information
  ROUTE TABLE:  1009 - 1764     -> Contains information about every route (36 routes x 21 bytes = 756 bytes)
  ROUTES:       1765 - 32004    -> Those memory addresses contains information about every route
                                One route: 40 coordinates x 21 bytes/coordinate = {40 x 21} = 840
                                36 routes: 36 x 840 = 30240
*/
  // sdaPin = Javelin's I/O pin P4, used for data
  // sclPin = Javelin's I/O pin P5, used for timing
  public MemoryRouteHandler( int sdaPin, int sclPin, GlobalVars vars)
  {

    m_eeprom = new MC24LC256(sdaPin, sclPin);    // Create MC24LC256 object

    // Other initializations
    m_globalVars = vars;
    //m_displayDriver = displayDriver;
  }
  public MC24LC256 GeteepromMemory() { return ( m_eeprom ); };

  /**
  Specification of one route (in route table):
  +-+---------------+
  |1|       20      |    (21 bytes): 1 byte is status -> 1 = used, 0 = unused; 20 bytes is the name of the route
  +-+---------------+
        ...
        ... (36 routes)
        ...
  +-+---------------+
  |1|       20      |
  +-+---------------+


  Specification of one route coordinate in the table (21 bytes)
  +-+------+------+------+
  |1| NAME | LAT  | LONG |    (21 bytes): 1 byte is status -> 1 = used, 0 = unused; NAME[8 bytes] = the name of the coordinate; LAT[6 bytes] = the latitude; LONG[6 bytes] = the longitude
  +-+------+------+------+
            ...
            ... (40 coordinates)
            ...
  +-+------+------+------+
  |1| NAME | LAT  | LONG |
  +-+------+------+------+

  REMARKS:
          How coordinates are represented:
          +---------+------------+----------------------+--------------------+----------------------+----------------------+--------------------+----------------------+
          | 1 byte  |  8 bytes   |        2 bytes       |     2 bytes        |      2 bytes         |        2 bytes       |     2 bytes        |      2 bytes         |
          +---------+------------+----------------------+--------------------+----------------------+----------------------+--------------------+----------------------+
            Status      Name         1 bit -> N/S          16 bits -> lat min   16 bits -> min dec     1 bit -> E/W          16 bits->long min    16 bits -> min dec
           1-used                   15 bit -> lat angle                                                15 bit -> long angle
           0-unused
  */

  private void SetCoordinate( int nAddress, int nDeg, int nMin, int nMinuteDec, boolean bFirstBit )
  {
    int nTemp = nDeg >> 8;
    nTemp |= ( bFirstBit ? 0x0080 : 0x0000 );

    // Write the degrees
    int nByte1 = nTemp;
    nByte1 &= 0x00FF;

    m_eeprom.writeVerify( m_dev, nAddress, (byte)nByte1 );

    int nByte2 = (nDeg & 0x00FF);
    nAddress += 1;
    m_eeprom.writeVerify( m_dev, nAddress, (byte)nByte2 );

    // Write the minutes
    nByte1 = nMin >> 8;
    nByte1 &= 0x00FF;
    nAddress += 1;
    m_eeprom.writeVerify( m_dev, nAddress, (byte)nByte1 );

    nByte2 = (nMin & 0x00FF);
    nAddress += 1;
    m_eeprom.writeVerify( m_dev, nAddress, (byte)nByte2 );

    // Write the minute decimal
    nByte1 = nMinuteDec >> 8;
    nByte1 &= 0x00FF;
    nAddress += 1;
    m_eeprom.writeVerify( m_dev, nAddress, (byte)nByte1 );

    nByte2 = (nMinuteDec & 0x00FF);
    nAddress += 1;
    m_eeprom.writeVerify( m_dev, nAddress, (byte)nByte2 );
  }

  private void GetCoordinate( int nAddress, GPSCoordinate gpsCoordinate ) throws MC24LC256BadReadException
  {
    // STEP 1:
    // Read the latitude degree
    int nTemp = m_eeprom.readRandom(m_dev, nAddress);
    if ((nTemp & 0x0080) > 0)
      gpsCoordinate.m_bNorth = true;

    gpsCoordinate.m_nLatDeg = (nTemp & 0x007F) << 8;
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    gpsCoordinate.m_nLatDeg |= nTemp;

    // Read the latitude minute
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    gpsCoordinate.m_nLatMinute = nTemp << 8;
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    gpsCoordinate.m_nLatMinute |= nTemp;

    // Read the latitude minute decimal
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    gpsCoordinate.m_nLatMinuteDec = nTemp << 8;
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    gpsCoordinate.m_nLatMinuteDec |= nTemp;

    // STEP 2:
    // Read the longitude degree
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    if ((nTemp & 0x0080) > 0)
      gpsCoordinate.m_bEast = true;

    gpsCoordinate.m_nLongDeg = (nTemp & 0x007F) << 8;
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    gpsCoordinate.m_nLongDeg |= nTemp;

    // Read the longitude minute
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    gpsCoordinate.m_nLongMinute = nTemp << 8;
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    gpsCoordinate.m_nLongMinute |= nTemp;

    // Read the longitude minute decimal
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    gpsCoordinate.m_nLongMinuteDec = nTemp << 8;
    nAddress += 1;
    nTemp = m_eeprom.readRandom(m_dev, nAddress);
    gpsCoordinate.m_nLongMinuteDec |= nTemp;

  }
  private void SetCoordinates( int nAddress, GPSCoordinate gpsCoordinate )
  {

    // Write the latitude
    SetCoordinate( nAddress,
                   gpsCoordinate.m_nLatDeg,         // The latitude degree
                   gpsCoordinate.m_nLatMinute,      // The latitude minute
                   gpsCoordinate.m_nLatMinuteDec,   // The latitude minute decimal
                   gpsCoordinate.m_bNorth );        // "true" if north

    nAddress += 6;
    // Write the longitude
    SetCoordinate( nAddress,
                   gpsCoordinate.m_nLongDeg,       // The longitude degree
                   gpsCoordinate.m_nLongMinute,    // The longitude minute
                   gpsCoordinate.m_nLongMinuteDec, // The longitude minute decimal
                   gpsCoordinate.m_bEast );        // "true" if east
  }
  private void SetCoordinateName( StringBuffer coordinateName, int nAddress )
  {
    // NOTE: The coordinate name is max 8 chars
    for ( int n = 0; n < 8; n++ )
    {
        int nCurrentAddress = (nAddress + n);

        char ch = 0;
        if (n < coordinateName.length())
          ch = coordinateName.charAt(n);

        m_eeprom.writeVerify( m_dev, nCurrentAddress, (byte)ch  ); // 1 -> indicates the coordinate is used
    }
  }

  public boolean IsRoute(int nRoute) throws MC24LC256BadReadException
  {
    if ((nRoute>36)||(nRoute==0)) return (false); // Check the boundary of the route. We can only have 36 routes...so addressing them goes from 0-35
    nRoute-=1; // Convert the route to an absolute place...

    return (m_eeprom.readRandom(m_dev, ROUTE_TABLE_START + 21*nRoute)==1);
  }
  // Function returns next available route slot in the EEPROM memory
  public int GetNextAvailableRouteSlot() throws MC24LC256BadReadException
  {
    for (int nRoute=0; nRoute<36; nRoute++)
    {
      if (!IsRoute(nRoute+1))
        return (nRoute+1);
    }

    // Indicates there are no more routes
    return (0);
  }
  // Function returns the name of the route "nRoute" and stores it in the "buffer"
  // The return value is "true" if the route could be found
  public boolean GetRouteName(int nRoute, StringBuffer buffer) throws MC24LC256BadReadException
  {
    if ((nRoute>36)||(nRoute==0)) return (false); // Check the boundary of the route. We can only have 36 routes...so addressing them goes from 0-35
    nRoute-=1; // Convert the route to an absolute place...

    // If this position is marked as "empty" then do not do anything more
    int nRoutePositionInTable = ROUTE_TABLE_START + (nRoute * 21);
    nRoutePositionInTable += 1; // Go past the status bit in the route table

    buffer.clear();

    // Read the route name
    for ( int n=0; n<20; n++ )
    {
      int nReadAddress = (nRoutePositionInTable+n);
      char ch = (char)m_eeprom.readRandom(m_dev, nReadAddress);
      buffer.append( ch );
    }

    return (true);
  }

  public boolean SetRouteName(int nRoute, StringBuffer buffer)
  {
    if ((nRoute>36)||(nRoute==0)) return (false); // Check the boundary of the route. We can only have 36 routes...so addressing them goes from 0-35
    nRoute-=1; // Convert the route to an absolute place...

    // If this position is marked as "empty" then do not do anything more
    int nRoutePositionInTable = ROUTE_TABLE_START + (nRoute * 21);
    nRoutePositionInTable += 1; // Go past the status bit in the route table

    int nBufferSize = buffer.length();

    // Read the route name
    for ( int n=0; n<20; n++ )
    {
      int nWriteAddress = (nRoutePositionInTable+n);
      int ch = (n>=nBufferSize) ? 0 : buffer.charAt(n);
      m_eeprom.writeVerify(m_dev, nWriteAddress, (byte)ch);
    }

    return (true);
  }

  // Finds the GPS coordinates at route "nRoute" and coordinate "nCoordinate"
  public boolean GetCoordinate( int nRoute, int nCoordinate, GPSCoordinate gpsCoordinate ) throws MC24LC256BadReadException
  {
    if ((nRoute>36)||(nRoute==0)) return (false); // Check the boundary of the route. We can only have 36 routes...so addressing them goes from 0-35
    if ((nCoordinate>40)||(nCoordinate==0)) return (false); // Check the boundary of the coordinate. We have like 40 coodinates in a route...so it goes from 0-39

    nRoute-=1;
    nCoordinate-=1;

    // If this position is marked as "empty" then do not do anything more
    int nRoutePositionInTable = ROUTE_TABLE_START + (nRoute * 21);
    if (m_eeprom.readRandom(m_dev, nRoutePositionInTable) == 0)
      return (false);

    int nRoutePosition = ROUTES_START + (840*nRoute);
    int nCoordinateStart = nRoutePosition + (nCoordinate*21);
    if (m_eeprom.readRandom(m_dev, nCoordinateStart) == 0) // Position should not 0
      return (false);

    StringBuffer buffer = m_globalVars.m_buffer1;
    buffer.clear();

    nCoordinateStart++; // Move to next position

    // STEP 1:
    // Read the coordinate name
    for ( int n=0; n<8; n++ )
    {
      int nReadAddress = (nCoordinateStart+n);
      char ch = (char)m_eeprom.readRandom(m_dev, nReadAddress);
      buffer.append( ch );
    }

    gpsCoordinate.SetCoordinateName(buffer.toString());

    // STEP 2:
    // Read the latitude and longitude
    nCoordinateStart += 8;
    GetCoordinate( nCoordinateStart, gpsCoordinate );

    return (true);
  }
  public void DeleteRoute( int nRoute )
  {
    if ((nRoute>36)||(nRoute==0)) return; // Check the boundary of the route. We can only have 36 routes...so addressing them goes from 0-35
    nRoute-=1;

    // Mark the position as "empty"
    int nRoutePositionInTable = ROUTE_TABLE_START + (nRoute * 21);
    m_eeprom.writeVerify(m_dev, nRoutePositionInTable, (byte)0);

    // Delete all the coodinates in the list
    int nRoutePosition = ROUTES_START + (840*nRoute);
    int nCoordinateStart = nRoutePosition;
    int nCoordinateEnd = nCoordinateStart + 840;
    for ( int n = nCoordinateStart; n < nCoordinateEnd; )
    {
      m_eeprom.writeVerify(m_dev, n, (byte)0 );
      n += 21;
    }
  }
  public int GetNumberOfCoordinatesInRoute( int nRoute ) throws MC24LC256BadReadException
  {
    if ((nRoute>36)||(nRoute==0)) return (0); // Check the boundary of the route. We can only have 36 routes...so addressing them goes from 0-35
    nRoute-=1;

    // Mark the position as "empty"
    int nRoutePositionInTable = ROUTE_TABLE_START + (nRoute * 21);
    if ( m_eeprom.readRandom(m_dev, nRoutePositionInTable) == 0 )
       return (0);

    int nNumberOfCoordinates = 0;
    int nRoutePosition = ROUTES_START + (840*nRoute);
    int nCoordinateStart = nRoutePosition;
    int nCoordinateEnd = nCoordinateStart + 840;
    for ( int n = nCoordinateStart; n < nCoordinateEnd; )
    {
      if ( m_eeprom.readRandom(m_dev, n) != 0 )
        nNumberOfCoordinates++;

      n += 21;
    }

    return (nNumberOfCoordinates);
  }
  // This function is returning the next available slot in the eeprom for storing the
  private int GetNextAvailableRouteCoordinateSlot(int nRoute) throws MC24LC256BadReadException
  {
    // Check the boundary for the "nRoute"
    if ((nRoute>36)||(nRoute==0)) return (0);
    nRoute-=1; // Convert the route to an absolute position

    // Find the start of the route in the route table
    int nRouteStart = ROUTES_START + (840*nRoute);

    // And this will be the end of the route
    int nRouteEnd = nRouteStart + 840;
    int nCoordinateAdd = 0;
    for (int n=nRouteStart; n<nRouteEnd;n+=21)
    {
      int nValue = m_eeprom.readRandom(m_dev, n);
      if ( nValue==0) // if this place is not used
        return (nCoordinateAdd);

      nCoordinateAdd++;
    }

    // Coming here, means we have an error
    return (nCoordinateAdd);
  }
  /*
  // This function is deleting a coordinate from the list. Since we're keeping the coordinate list sorted, when we delete a coordinate
  // from a route, we move all the items in the list one step from the left
  public boolean DeleteCoordinate(int nRoute, int nPosition) throws MC24LC256BadReadException
  {
    // Check the boundary of the route. We can only have 36 routes...so addressing them goes from 0-35
    if ((nRoute>36)||(nRoute==0)) return (false);
    if ((nPosition>40)||(nPosition==0)) return (false);

    // Clear the first and the second row
    //m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 0, 1, 0, 20 );
    //m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 1, 2, 0, 20 );

    //m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 0, "Removing coordinate" );
    //m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 1, 0, "Working"  );

    int nNextSlot = GetNextAvailableRouteCoordinateSlot(nRoute);
    if (nNextSlot==0) // Cannot insert a coordinate if there are none in the list
      return (false);

    nPosition-=1; // Convert the position to an absolute position (the position is something like: 1, 2, 3...40 - never 0)
    nRoute-=1;    // Convert the route to an absolute position in the route table
    nNextSlot-=1;  // This is the last object in the list
    if (nPosition>nNextSlot)
      return (false);

    int nCoordinateDeletePositionInTable=ROUTES_START + (840*nRoute) + (21 * nPosition);
    int nLastCoordinatePositionInTable=ROUTES_START + (840*nRoute) + (21 * nNextSlot);
    nCoordinateDeletePositionInTable+=21;

    GPSCoordinate gpsCoordinate = m_globalVars.m_gpsCoordinate1;
    int nPos=nPosition+1;

    for (int n=nCoordinateDeletePositionInTable; n<=nLastCoordinatePositionInTable; n+=21)
    {

      //int nScreenPosition = m_displayDriver.GetLastPosition(DisplayDriver.DISPLAY_1);
      //m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1,
      //                            (nScreenPosition >> 8),
      //                            ((nScreenPosition & 0x00FF)+1),
      //                            "." );


      // ...get the coordinate at place nPosition+1
      GetCoordinate((nRoute+1), (nPos+1), gpsCoordinate);

      // ...and move it to position nPos
      if (!SaveCoordinateToEEPROM(gpsCoordinate, (n-21)))
        return (false);

      nPos++;
    }
    // Mark the last position in table as not used
    m_eeprom.writeVerify(m_dev, nLastCoordinatePositionInTable, (byte)0 );

    return (true);
  }

  // This function is inserting a coordinate in the slot position and returns "true" if the inserting was OK.
  // The only reason why this function might fail is because or all slots beeing used or if the "nPosition" is greater
  // than "nNextSlot"
  public boolean InsertCoordinate(int nRoute, int nPosition, GPSCoordinate gpsInsertCoordinate) throws MC24LC256BadReadException
  {
    // Check the boundary of the route. We can only have 36 routes...so addressing them goes from 0-35
    if ((nRoute>36)||(nRoute==0)) return (false);
    if ((nPosition>40)||(nPosition==0)) return (false);

    // Clear the first and the second row
    //m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 0, 1, 0, 20 );
    //m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 1, 2, 0, 20 );

    //m_displayDriver.WriteMessage( DisplayDriver.DISPLAY_1, 0, 0, "Inserting coordinate" );
    //m_displayDriver.WriteMessage( DisplayDriver.DISPLAY_1, 1, 0, "Working" );

    // NOTE!
    // When inserting a coordinate in the list, we move all the coordinates after the insertion slot
    // one step backward. This is a time consuming procedure but the advantage is that we always know that
    // the coordinates in the list are sorted correctly. Insertion of coordinates in the list is only done
    // sometimes
    int nNextSlot = GetNextAvailableRouteCoordinateSlot(nRoute);
    if (nNextSlot==0) // Cannot insert a coordinate if there are none in the list
      return (false);

    nPosition-=1; // Convert the position to an absolute position (the position is something like: 1, 2, 3...40 - never 0)
    nRoute-=1;    // Convert the route to an absolute position in the route table
    nNextSlot-=1;  // This is the last object in the list
    if (nPosition>nNextSlot)
      return (false);

    //int nRoutePositionInTable = ROUTE_TABLE_START + (nRoute * 21);
    int nCoordinateInsertionPositionInTable=ROUTES_START + (840*nRoute) + (21 * nPosition);
    int nLastCoordinatePositionInTable=ROUTES_START + (840*nRoute) + (21 * nNextSlot);

    // Move all the items from nLastCoordinatePositionInTable to nCoordinatePositionInTable one step
    int nItem = nLastCoordinatePositionInTable;
    GPSCoordinate gpsCoordinate = m_globalVars.m_gpsCoordinate1;
    while (nItem>=nCoordinateInsertionPositionInTable)
    {
      int n = nItem;

      //int nScreenPosition = m_displayDriver.GetLastPosition(DisplayDriver.DISPLAY_1);
      //m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1,
      //                            (nScreenPosition >> 8),
      //                            ((nScreenPosition & 0x00FF)+1),
      //                            "." );


      // Get the coordinate at nItem...
      GetCoordinate((nRoute+1), (nNextSlot+1), gpsCoordinate);

      // ...and move it at nItem+1
      n+=21; // -> move to next coordinate in list

      // Let this function save the coordinate to the EEPROM
      if (!SaveCoordinateToEEPROM(gpsCoordinate, n))
        return (false);

      nItem-=21;
      nNextSlot-=1;
    }

    // Finally save the coordinate to be inserted in the right spot
    if (!SaveCoordinateToEEPROM(gpsInsertCoordinate, nCoordinateInsertionPositionInTable))
       return (false);


    return (false);
  }
  */
  // This function takes a gps coordinate and adds saves it to the "nPlace". There is no checking in this function
  private boolean SaveCoordinateToEEPROM(GPSCoordinate gpsCoordinate, int nPlace)
  {
      // 1 -> indicates the coordinate is used
      byte myByte = (byte)1;
      myByte &= 0x00FF;
      m_eeprom.writeVerify(m_dev, nPlace, myByte);

      // Coordinate name
      nPlace += 1;
      SetCoordinateName(gpsCoordinate.GetCoordinateName(), nPlace);

      // LAT/LONG
      nPlace += 8;
      SetCoordinates(nPlace, gpsCoordinate);

    return (true);
  }

  // Adding a coordinate adds a coordinate to the route "nRoute". The new coordinate is always added to the end of the list. We have just about 40 coordinates
  // so, if the function can not find a place in the table "0" will be returned, in other case the current place in the table where the coordinate where added is returned
  public int AddCoordinate(int nRoute, GPSCoordinate gpsCoordinate ) throws MC24LC256BadReadException
  {
    // Check the boundary of the route. We can only have 36 routes...so addressing them goes from 0-35
    if ((nRoute>36)||(nRoute==0)) return (0);
    nRoute-=1; // Convert the route position to an absolute position

    // Clear the first and the second row
    //m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 0, 1, 0, 20 );
    //m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 1, 2, 0, 20 );

    //m_displayDriver.WriteMessage( DisplayDriver.DISPLAY_1, 0, 0, "Adding coordinate" );
    //m_displayDriver.WriteMessage( DisplayDriver.DISPLAY_1, 1, 0, "Working" );

    // STEP 1:
    // First make sure the route in the route table is marked as used
    int nRoutePositionInTable = ROUTE_TABLE_START + (nRoute * 21);
    m_eeprom.writeVerify(m_dev, nRoutePositionInTable, (byte)1); // 1 -> indicates the route is used

    // STEP 2:
    // Find the start of the route
    int nRouteStart = ROUTES_START + (840 * nRoute);

    // STEP 3:
    // Find the next available coordinate in the table
    int nPlace = 0;
    nPlace=GetNextAvailableRouteCoordinateSlot(nRoute+1);
    if (nPlace>=40) // Boundary check
      return (0);

    int n=(ROUTES_START + (840*nRoute)+ (nPlace*21));

    // STEP 4:
    // Store the coordinate
    {
      // Let this function save the coordinate to the EEPROM
      if (!SaveCoordinateToEEPROM(gpsCoordinate, n))
        return (0);
    }
    return (nPlace+1);
  }
  public boolean IsFormatted() // PE.060424. No more space
  {
    return true;
  }
  /*
  // Function returns true if the MC24LC256 needs to be formatted
  public boolean IsFormatted()
  {
    boolean b = true;
    try
    {
      // STEP 1:
      // ---------------------------------------------------------------
      // Read the status byte
      b &= ( (byte)m_eeprom.readRandom( m_dev, FOMMAT_TABLE_START ) == (byte)1 );

      // STEP 2:
      // ---------------------------------------------------------------
      // Read the LSB
      int i = (m_eeprom.readRandom( m_dev, (FOMMAT_TABLE_START+1) ) & 0x00FF );
      int nSize = i;

      // Read the MSB
      i = m_eeprom.readRandom( m_dev, (FOMMAT_TABLE_START+2) ) << 8;
      nSize |= i;

      b &= ( nSize == m_eepromSize ); // Check if the size is the same

      StringBuffer buffer = m_globalVars.m_buffer1;
      buffer.clear();

      // STEP 3:
      // ---------------------------------------------------------------
      // Read the text "OK"
      buffer.append( (char)m_eeprom.readRandom( m_dev, (FOMMAT_TABLE_START+3) ) );
      buffer.append( (char)m_eeprom.readRandom( m_dev, (FOMMAT_TABLE_START+4) ) );

      String str = buffer.toString();
      b &= str.equalsIgnoreCase("OK");

    }
    catch( Exception ex ) { b = false; }

    return (b);
  }

  // Function is adding a header to the beginning of the memory block
  private boolean WriteFormatPage()
  {
    // STEP 1:
    // ---------------------------------------------------------------
    m_eeprom.writeVerify( m_dev, FOMMAT_TABLE_START, (byte)1 ); // 1 -> indicates the unit is formatted [1 byte]

    // 32000 = 0x7D00

    // STEP 2:
    // ---------------------------------------------------------------
    // LSB
    int iTemp = (m_eepromSize & 0x00FF);
    m_eeprom.writeVerify( m_dev, (FOMMAT_TABLE_START+1), (byte)iTemp ); // FF -> write the lower size bits [1 byte]

    // STEP 3:
    // ---------------------------------------------------------------
    // MSB
    iTemp = ((m_eepromSize & 0xFF00) >> 8 );
    m_eeprom.writeVerify( m_dev, (FOMMAT_TABLE_START+2), (byte)iTemp ); // 7F -> write the lower size bits [1 byte]

    StringBuffer buffer = m_globalVars.m_buffer1;
    buffer.clear();
    buffer.append("OK");
    m_eeprom.writeVerify( m_dev, (FOMMAT_TABLE_START+3), (byte)buffer.charAt(0) );
    m_eeprom.writeVerify( m_dev, (FOMMAT_TABLE_START+4), (byte)buffer.charAt(1) );

    // STEP 4:
    // Initialize all the routes
    ClearAllRoutes();

    return (true);
  }
  */

  /**
    Specification of one route (in route table):
    +-+---------------+
    |1|       20      |    (21 bytes): 1 byte is status -> 1 = used, 0 = unused; 20 bytes is the name of the route
    +-+---------------+
          ...
          ... (36 routes)
          ...
    +-+---------------+
    |1|       20      |
    +-+---------------+
  */

  public boolean ClearAllRoutes() { return InitalizeRouteTable(); };
  public boolean ClearRoute( int nRoute )
  {
    if ((nRoute>36)||(nRoute==0)) return (false);
    nRoute-=1; // Convert the route to an absolute route

    // STEP 1:
    // First, clear the route in the route table
    int nTableRouteStart = ROUTE_TABLE_START + (nRoute * 21); //(756 = 36*21)
    m_eeprom.writeVerify( m_dev, nTableRouteStart, (byte)0 );

    // STEP 2:
    // Clear the actual route
    InitializeRoute( nRoute );

    return (true);
  }

  private boolean InitalizeRouteTable()
  {
    // The route table starts at "ROUTE_TABLE_START"
    int nRoutesTableEnd = ROUTE_TABLE_START + 756; //(756 = 36*21)

    for ( int n = ROUTE_TABLE_START; n < nRoutesTableEnd; )
    {
      m_eeprom.writeVerify( m_dev, n, (byte)0 );
      n += 21; // Move to next position
    }

    InitializeRoutes();
    return (true);
  }

  // This function initializes the routes - it ensures all the routes are zeroed
  // We have 36 routes
  private boolean InitializeRoutes()
  {

    //int nScreenPosition = m_displayDriver.GetLastPosition(DisplayDriver.DISPLAY_1);
    //StringBuffer buffer = m_globalVars.m_buffer1;
    for ( int nRoute = 0; nRoute < 36; nRoute++ )
    {
      /*
      m_displayDriver.ClearScreenRange(DisplayDriver.DISPLAY_1, 3, 4, 12, 20 );
      buffer.clear();
      buffer.append(nRoute+1);
      buffer.append("/");
      buffer.append(36);
      m_displayDriver.WriteMessage( DisplayDriver.DISPLAY_1,
                                  (nScreenPosition >> 8),
                                  ((nScreenPosition & 0x00FF)+1),
                                   buffer.toString() );
      */
      InitializeRoute(nRoute+1);
    }
    return (true);
  }
  // Function zeroes a route - a route contains 40 positions, or coordinates
  /**
  +-+------+------+------+
  |1| NAME | LAT  | LONG |    (21 bytes): 1 byte is status -> 1 = used, 0 = unused; NAME = the name of the coordinate; LAT = the latitude; LONG = the longitude
  +-+------+------+------+
            ...
            ... (40 coordinates)
            ...
  +-+------+------+------+
  |1| NAME | LAT  | LONG |
  +-+------+------+------+
  */
  private boolean InitializeRoute( int nRoute )
  {
    if ((nRoute>36)||(nRoute==0)) return (false);
    nRoute-=1; // Convert the route to an absolute route

    // The start address is the start of the "routes" tables + the size of one route table * the current route
    int nStartAddress = ROUTES_START + (840 * nRoute);
    int nRouteEnd = (nStartAddress + 840);

    StringBuffer buffer = m_globalVars.m_buffer1;

    // Clear all the coordinates
    int iPos = 0;
    for ( int nPosition = nStartAddress; nPosition < nRouteEnd; nPosition += 21 )
    {
      /*
      buffer.clear();
      buffer.append("->Position:");
      buffer.append(iPos+1);
      buffer.append("/");
      buffer.append(40);
      m_displayDriver.WriteMessage( DisplayDriver.DISPLAY_1,
                                    3, // Last row
                                    0,
                                    buffer.toString() );
      */

      m_eeprom.writeVerify( m_dev, nStartAddress + nPosition, (byte)0 );
      iPos++;
    }
    return (true);
  }
  // Function is "formatting" the eeprom memory.
  public boolean Format()
  {
    //return (WriteFormatPage()); // PE.060424. No more space
    return true; // PE.060424. No more space
  }

}