
package stamp.peripheral.sensor.GPS.GPSMath;

import stamp.peripheral.sensor.GPS.NMEA.*;
import stamp.core.*;
import com.micromegacorp.math.hardware.*;
import stamp.peripheral.sensor.GPS.StringOperations.*;  // Contains some string handling classes
import stamp.peripheral.sensor.GPS.Include.*;           // Contains the GPS message list and more

// This class contains functions used to parse through the messages
// we receive from the GPS module
public class NMEA0183
{
  private StringBuffer m_strTemp;
  //private GPSMessageList m_messageListGGA = null;
  private GPSMessageList m_messageListVTG;
  private ZDA m_MessageZDA;
  private GGA m_MessageGGA;
  private GlobalVars m_globalVars;
  private Uart m_rxGPS;
  private StringBuffer m_buffer;
  public static final int GPS_MESSAGE_BUFFER = 165; // ~the length of ZDA + GGA + VTG

  public NMEA0183(Uart rxGPS, GlobalVars globalVars)
  {
    m_globalVars = globalVars;

    // Create the lists
    //m_messageListGGA = new GPSMessageList(m_globalVars);
    m_messageListVTG = new GPSMessageList();

    //m_messageListGGA.CreateMessageList(NMEADefinitions.GGA_MESSAGE, 5);
    m_messageListVTG.CreateMessageList(NMEADefinitions.VTG_MESSAGE, 5);

    m_strTemp = m_globalVars.m_buffer1;
    m_buffer = new StringBuffer(GPS_MESSAGE_BUFFER);

    m_MessageZDA = new ZDA(m_globalVars.m_timeParser);
    m_MessageGGA = new GGA(m_globalVars.m_timeParser);
    m_rxGPS = rxGPS; // This is the class listening to incomming GPS messages
  }
  //public GGA GetGGAMessage() { return ( (GGA)(m_messageListGGA.GetCurrentMessage())); };
  public GGA GetGGAMessage() { return ( m_MessageGGA ); };
  public VTG GetVTGMessage() { return ( (VTG)(m_messageListVTG.GetCurrentMessage() )); };
  public ZDA GetZDAMessage() { return ( m_MessageZDA ); };

  // SMG
  // Speed Made Good returns the mean value of all the readings...(only the valid ones)
  public boolean GetSpeedMadeGood( Float32 fTemp )
  {
    fTemp.set(0);
    VTG vtgCurrent = (VTG)(m_messageListVTG.GetCurrentMessage());
    if (vtgCurrent == null) return false;

    int nCount = 0;
    if ( vtgCurrent.GetIsValid() )
    {
      fTemp.set(vtgCurrent.m_nSpeedKnots);
      nCount = 1;
    }
    VTG vtgPrev = (VTG)vtgCurrent.GetPrevItem();
    while ( vtgPrev != vtgCurrent )
    {
      if (vtgPrev.GetIsValid())
      {
        fTemp.add(vtgPrev.m_nSpeedKnots);
        nCount++;
      }
      vtgPrev = (VTG)vtgPrev.GetPrevItem();
    }

    if (nCount==0) return (false);

    fTemp.divide( nCount );
    fTemp.divide( 100 );

    return (true);
  }

  public int Process()
  {
    if (m_buffer.length()>0)
    //if (GetMessage(m_rxGPS, m_buffer))
      return ( ParseGPSMessage(m_buffer ));

    return (NMEADefinitions.OOO_MESSAGE);
  }
  private int ParseGPSMessage(StringBuffer strMessage)
  {
    m_MessageZDA.Clear();
    m_strTemp.clear();
    return ( ParseNMEA_Message( strMessage ) );
  }

  /**
     This function parses one or several NMEA 0183 messages and returns the message we've manage to parse
     please see the definitions above
     The "buffer" contains a list of NMEA0183 messages in a queue, i.e. the most recent message is added at the back of the message (string)...

     Example:
     $GPGGA,210006.00,5535.4468,N,01300.3073,E,1,03,3.99,00035,M,040,M,,*66CR/LF$GPGGA,,,,,,0,,,,,040,M,,*66CR/LF$GPVTG,89.68,T,,M,0.00,N,0.0,K*5FCR/LF$GPVTG,89.68,T,,M,0.00,N,0.0,K*5FCR/LF
     |--------------------------------1-----------------------------------|     |---------------2----------|     |------------3------------------|     |-------------4-----------------|

     In the message above there are 4 messages. "3" and "4" have the same ID and "1" and "2" have also the same ID...
     but message "2" is not valid, while message "3" and "4" are the same

  */
  public int ParseNMEA_Message( StringBuffer buffer )
  {
    int nMessages = NMEADefinitions.OOO_MESSAGE; // There is no message to start with

    int nLastPos = ( buffer.length() - 1 );
    if ( nLastPos <= 0 )
      return ( nMessages );

    int nMessageType = NMEADefinitions.OOO_MESSAGE;

    // Loop through all available messages
    while ( ( nLastPos = StringOperations.GetGPSMessage( buffer, 0, nLastPos, m_strTemp ) ) >= 0 )
    {
      nMessageType = GetMessageType( m_strTemp ); // Get the message type
      nLastPos -= 1;

      switch( nMessageType )
      {
        case NMEADefinitions.GGA_MESSAGE: // GGA message
        {

          if ( ( nMessages & NMEADefinitions.GGA_MESSAGE ) > 0 ) break; // No need to process the same message twice
          if ( m_MessageGGA.Parse( m_strTemp, m_globalVars.m_fTemp1 ) )
            nMessages |= NMEADefinitions.GGA_MESSAGE; // There is aa GGA message available
        }
        break;
        case NMEADefinitions.VTG_MESSAGE: // VTG Message
        {
          if ( ( nMessages & NMEADefinitions.VTG_MESSAGE ) > 0 ) break; // No need to process the same message twice

          VTG vtg = (VTG)m_messageListVTG.MoveForwardToNextMessage();
          if ( vtg != null )
          {
            if ( vtg.Parse( m_strTemp, m_globalVars.m_fTemp1 ) )
              nMessages |= NMEADefinitions.VTG_MESSAGE; // There is a VTG message available
            else
            {
              // Move to previous message
              m_messageListVTG.MoveBackwardToPreviousMessage();
            }
          }
        }
        break;
        case NMEADefinitions.ZDA_MESSAGE: // ZDA Message
        {
          if ( ( nMessages & NMEADefinitions.ZDA_MESSAGE ) > 0 ) break; // No need to process the same message twice
          if ( m_MessageZDA.Parse( m_strTemp ) )
            nMessages |= NMEADefinitions.ZDA_MESSAGE; // There is a ZDA message available
        }
        break;
      }
    }
    return ( nMessages );
  }
  public static int GetMessageType( StringBuffer strBuffer )
  {
    int nMessage = NMEADefinitions.OOO_MESSAGE;
    if (strBuffer.indexOf("$GPGGA") >=0 )
      return ( NMEADefinitions.GGA_MESSAGE );

    /*
    if (strBuffer.indexOf("$GPGSA") >=0 )
      return ( NMEADefinitions.GSA_MESSAGE );

    if (strBuffer.indexOf("$GPRMC") >=0 )
      return (NMEADefinitions.RMC_MESSAGE );
    */
    if (strBuffer.indexOf("$GPVTG") >=0 )
      return ( NMEADefinitions.VTG_MESSAGE );

    if (strBuffer.indexOf("$GPZDA") >=0 )
      return ( NMEADefinitions.ZDA_MESSAGE );

    return ( NMEADefinitions.OOO_MESSAGE );
  }
  // Used to dump the current GPS messages ...
  public void CaptureGPSFeed(boolean bClear)
  {
    if (bClear) m_buffer.clear();
    GetMessage(m_rxGPS, m_buffer);
  }
  public int GetGPSBufferLength()
  {
    return m_buffer.length();
  }
  private boolean GetMessage( Uart rxGPS, StringBuffer buffer )
  {
    // Make sure the buffer lenght is OK
    if ( ( buffer.length() >= GPS_MESSAGE_BUFFER - 1 ) ) return true;

    char ch = '0';
    boolean b = false;
    while ( rxGPS.byteAvailable() )
    {
      ch = (char)rxGPS.receiveByte();
      if ( ( buffer.length() >= GPS_MESSAGE_BUFFER - 1 ) )
      {
        // Empty the buffer
        while ( rxGPS.byteAvailable() )
        {
          rxGPS.receiveByte();
        }
      }

      buffer.append( ch );
      b = true;
    }
    return ( b );
  }
}