
package stamp.peripheral.sensor.GPS.Kernel;

import stamp.core.*;
import stamp.math.*;                                    // Contains the uM-FPU implementation

import stamp.peripheral.memory.eeprom.*;                // Contains handling for EEPROM operations
import stamp.peripheral.sensor.GPS.GPSMath.*;           // Contains the GPS calculations
import stamp.peripheral.sensor.GPS.NMEA.*;              // Contains the NMEA messages and other NMEA specific things
import stamp.peripheral.sensor.GPS.StringOperations.*;  // Contains some string handling classes
import stamp.peripheral.sensor.GPS.Include.*;           // Contains the linked list class and other stuff
import stamp.peripheral.sensor.GPS.Keyboard.*;          // Contains the memkey implementation
import stamp.peripheral.sensor.GPS.Temperature.*;       // Contains the SHT11 sensor wrapper
import stamp.peripheral.sensor.GPS.DisplayDriver.*;     // Contains the LCD driver for the CrystalFontz
import stamp.peripheral.sensor.GPS.RouteFollower.*;     // Contains the class handling keeping track of were we are and where to go
import stamp.peripheral.sensor.GPS.Menu.*;              // Contains the GPS menu

// This class is the actual GPS engine
public class Kernel
{
  public static final int KERNEL_MODE_EXIT=0;  // This mode will exit the GPS
  public static final int GPS_MODE_INIT=1;     // Mode after the first initialization done in the "SailSmart" class
  public static final int GPS_MODE_RUN=2;      // Kernel is running
  public static final int GPS_MODE_IDLE=3;     // In the IDLE mode, the GPS does not do anything, but listens to the keyboard

  private static final int KERNEL_PROCESS_NONE                 = 0x0000;     // Kernel should not process anything
  private static final int KERNEL_PROCESS_REFRESH_TEMPERATURE  = 0x0001;     // Read and show the temperature
  private static final int KERNEL_PROCESS_READ_GPS_MESSAGE     = 0x0002;     // Read the GPS message
  private static final int KERNEL_RROCESS_SHOW_GPS_POSITION    = 0x0004;     // Show the current GPS position
  private static final int KERNEL_PROCESS_CALC_MEAN_COURSE     = 0x0008;     // Calculate the mean course
  private static final int KERNEL_PROCESS_DISPLAY_IDLE_ICON    = 0x0010;     // Displays the idle icon
  private static final int KERNEL_PROCESS_FOLLOW_ROUTE         = 0x0020;     // Route follower


  private static final int KERNEL_PROCESS_ALL = ( KERNEL_PROCESS_REFRESH_TEMPERATURE |
                                                  KERNEL_PROCESS_READ_GPS_MESSAGE |
                                                  KERNEL_RROCESS_SHOW_GPS_POSITION |
                                                  KERNEL_PROCESS_CALC_MEAN_COURSE |
                                                  KERNEL_PROCESS_DISPLAY_IDLE_ICON |
                                                  KERNEL_PROCESS_FOLLOW_ROUTE );

  private int m_KernelMode = KERNEL_MODE_EXIT;

  public DisplayDriver m_displayDriver;       // Class used for handling the display
  public GlobalVars m_globalVars;             // Contains global variables used in the whole program
  MemoryRouteHandler m_routeHandler;   // The route handler (which is a wrapper for the EEPROM) contains all the routes
  NMEA0183 m_NMEA0183;
  MEMKey m_keyboard;
  Temperature m_Temperature;
  RouteFollower m_RouteFollower;
  GPSMath m_gpsMath;
  Timer m_timer;
  GPSMenuHandler m_menuHandler;
  Settings m_settings;
  BacklightHandler m_backlightHandler;

  public Kernel(DisplayDriver displayDriver, GlobalVars globalVars, MemoryRouteHandler routeHandler, NMEA0183 nmea0183, MEMKey keyboard, Temperature temp, RouteFollower routeFollower, GPSMath gpsMath)
  {
    m_timer = new Timer();
    m_KernelMode=GPS_MODE_INIT;

    m_displayDriver=displayDriver;
    m_globalVars=globalVars;
    m_routeHandler=routeHandler;
    m_NMEA0183=nmea0183;
    m_keyboard=keyboard;
    m_Temperature=temp;
    m_RouteFollower = routeFollower;
    m_gpsMath = gpsMath;
    m_settings = new Settings();
    m_menuHandler= new GPSMenuHandler(displayDriver, globalVars, routeHandler, keyboard, routeFollower, gpsMath, nmea0183, m_settings);

    // Read all the settings from eeprom
    m_RouteFollower.SetSettings(m_settings);
    m_settings.ReadSettings(m_routeHandler.GeteepromMemory());
    m_RouteFollower.SetStoredSettings();
    m_backlightHandler = new BacklightHandler();
  }

  public void SetKernelMode(int nMode){ m_KernelMode=nMode;};
  public int GetKernelMode() {return (m_KernelMode);};

  // This function is the main execution loop
  public void Go()
  {
    m_timer.mark();
    int nKernel = KERNEL_PROCESS_NONE;
    int nCaptureGPSFeed = 0;
    while (m_KernelMode!=KERNEL_MODE_EXIT)
    {
      switch( m_KernelMode )
      {
        case GPS_MODE_INIT:
        {
          if (m_timer.timeout(2000))
          {
             m_displayDriver.ClearScreen(DisplayDriver.DISPLAY_1 | DisplayDriver.DISPLAY_2);
             m_displayDriver.ShowCursor(DisplayDriver.DISPLAY_1|DisplayDriver.DISPLAY_2, false);
             m_timer.mark();
             m_KernelMode=GPS_MODE_RUN; // Switch to running state
             nKernel = KERNEL_PROCESS_ALL;
          }
        }
        break;

        case GPS_MODE_RUN:
        {

          /**********************************************************************************/

          /**********************************************************************************/
          /**********|                   Toggle the icon              |**********************/
          /**********************************************************************************/
          if (m_timer.timeout(1500) && ( nKernel & KERNEL_PROCESS_DISPLAY_IDLE_ICON ) > 0 )
          {
            m_displayDriver.WriteCharacter(DisplayDriver.DISPLAY_1, 0, 19, (byte)DisplayDriver.SPECIAL_CHARACTER_HEART_BEAT );
            nKernel &= ~KERNEL_PROCESS_DISPLAY_IDLE_ICON;
          }
          else if ( m_timer.timeout(3000) && ( nKernel & KERNEL_PROCESS_DISPLAY_IDLE_ICON ) == 0)
          {
            m_displayDriver.WriteCharacter(DisplayDriver.DISPLAY_1, 0, 19, (byte)' ' );
          }
          /**********************************************************************************/

          // Let the "Route Follower" refresh itself every 2 seconds
          if ( m_timer.timeout(2000) && (nKernel & KERNEL_PROCESS_FOLLOW_ROUTE ) > 0 )
          {
            GGA gga = m_NMEA0183.GetGGAMessage();
            if (gga != null )
            {
              GPSCoordinate lastCoordinate = gga.GetGPSCoordinate();
              m_RouteFollower.Update( lastCoordinate, m_gpsMath, m_displayDriver, m_globalVars, m_NMEA0183 );
              nKernel &= ~KERNEL_PROCESS_FOLLOW_ROUTE;
            }
          }
          if ( m_timer.timeout(2000) && !( m_timer.timeout(4000) ) )
          {
            m_backlightHandler.Idle(m_globalVars.m_fTemp1);
          }
          else if ( m_timer.timeout(4000) )
            m_backlightHandler.Idle(m_globalVars.m_fTemp1);

          MEMKeyInput input = m_keyboard.GetCurrentChar(MEMKey.KEYBOARD_MODE_EXTRA);
          switch ( input.m_Key )
          {
            case MEMKey.KEYBOARD_BYTE_D:
            {
              // Try to see if the user pushed "D"
              m_menuHandler.RunMenu();
              m_KernelMode=GPS_MODE_INIT; // Switch to init state
              continue;
            }
            case MEMKey.KEYBOARD_BYTE_A:
            {
              // Try to see if the user pushed "A" button - which means we goto next waypoint
              m_RouteFollower.HandleGotoNextWaypoint(m_displayDriver);
              //continue;
            }
            break;
            case MEMKey.KEYBOARD_BYTE_B:
            {
              // Try to see if the user pushed "A" button - which means we goto previous waypoint
              m_RouteFollower.HandleGotoPreviousWaypoint(m_displayDriver);
              //continue;
            }
            break;

            case MEMKey.KEYBOARD_BYTE_C:
            {
              m_RouteFollower.HandleDisconnectFromRoute(m_displayDriver);
              //continue;
            }
            break;
            default: break;
          }

          // Refresh temperature every 1.5 seconds
          if ( m_timer.timeout(1500) && (nKernel & KERNEL_PROCESS_REFRESH_TEMPERATURE ) > 0 )
          {
            DisplayTemperature(DisplayDriver.DISPLAY_2);
            nKernel &= ~KERNEL_PROCESS_REFRESH_TEMPERATURE;

            /*
            displayDriver.ShowCursor(DisplayDriver.DISPLAY_1, true);
            m_globalVars.m_buffer1.clear();
            m_displayDriver.ClearScreenRange( DisplayDriver.DISPLAY_2, 1, 2, 0, 20 );
            MEMKeyInput input = m_keyboard.GetChar(true, MEMKey.KEYBOARD_MODE_ENTER_TEXT, m_displayDriver, DisplayDriver.DISPLAY_2, 1, 0, m_globalVars.m_buffer1, 20, MEMKey.KEYBOARD_BYTE_D, MEMKey.KEYBOARD_BYTE_A, MEMKey.KEYBOARD_BYTE_B, MEMKey.KEYBOARD_BYTE_C );

            m_displayDriver.ClearScreenRange( DisplayDriver.DISPLAY_2, 1, 2, 0, 20 );
            m_displayDriver.WriteMessage( DisplayDriver.DISPLAY_2, 2, 0, m_globalVars.m_buffer1.toString() );
            */

          }

          // Start collecting the messages first after 2 secs
          if (m_timer.timeout(2000)) m_NMEA0183.CaptureGPSFeed(false);

          if ( ( m_NMEA0183.GetGPSBufferLength() > 120 ) && (nKernel & KERNEL_PROCESS_READ_GPS_MESSAGE ) > 0 )
          {
            // Process the GPS messages
            int nProcess = m_NMEA0183.Process();

            // Is there any ZDA message?
            if ( ( nProcess & NMEADefinitions.ZDA_MESSAGE ) > 0 )
            {
              // Show the time...
              m_NMEA0183.GetZDAMessage().GetUTC( m_globalVars.m_buffer1 );
              m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 3, 0, m_globalVars.m_buffer1.toString() );

              m_NMEA0183.GetZDAMessage().GetDate( m_globalVars.m_buffer1 );
              m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_2, 3, 6, m_globalVars.m_buffer1.toString() );
            }

            // Is there any GGA message?
            if ( ( nProcess & NMEADefinitions.GGA_MESSAGE ) > 0 )
            {
              m_NMEA0183.GetGGAMessage().GetLatitude( m_globalVars.m_buffer1 );
              m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 0, 1, m_globalVars.m_buffer1.toString() );

              m_NMEA0183.GetGGAMessage().GetLongitude( m_globalVars.m_buffer1 );
              m_displayDriver.WriteMessage(DisplayDriver.DISPLAY_1, 1, 1, m_globalVars.m_buffer1.toString() );
            }

            // Show the antenna symbol (this means we get info from the IQ Lassen)
            m_displayDriver.WriteCharacter(DisplayDriver.DISPLAY_1, 0, 18, ( nProcess == NMEADefinitions.OOO_MESSAGE ) ? (byte)' ' : (byte)DisplayDriver.SPECIAL_CHARACTER_GPS_COVERAGE );
            nKernel &= ~KERNEL_PROCESS_READ_GPS_MESSAGE;
          }

          /**********************************************************************************/
          /**********|                   MOB                          |**********************/
          /**********************************************************************************/
          input = m_keyboard.GetCurrentChar(MEMKey.KEYBOARD_MODE_SPECIAL);
          switch ( input.m_Key )
          {
            case MEMKey.KEYBOARD_BYTE_STAR:
            {
              GGA gga = m_NMEA0183.GetGGAMessage();
              if (gga != null )
              {
                GPSCoordinate lastCoordinate = gga.GetGPSCoordinate();

                m_RouteFollower.ToggleMobState(lastCoordinate, m_displayDriver);
              }
              break;
            }
          }
          /**********************************************************************************/
          /**********|                   MOB                          |**********************/
          /**********************************************************************************/


          // Restart the timer from this time position (every 6 secs)
          if (m_timer.timeout(6000))
          {
            m_timer.mark();
            m_NMEA0183.CaptureGPSFeed(true);

            // Restore everything
            nKernel = KERNEL_PROCESS_ALL;
            nCaptureGPSFeed = 2000;
          }
        }
        break;

      }
    }
  }

  private void DisplayTemperature(int nDisplay)
  {
    m_globalVars.m_buffer1.clear();
    m_Temperature.MeasureTemp(m_globalVars.m_buffer1);
    m_displayDriver.WriteMessage(nDisplay, 3, 13, (m_globalVars.m_buffer1).toString());
  }
}